/***********************
@by: Wang Cheng Wei
@2020.12
@reference resources :
@https://blog.csdn.net/sandalphon4869/article/details/103841275?utm_medium=distribute.pc_relevant.none-task-blog-title-2&spm=1001.2101.3001.4242
@https://blog.csdn.net/lemonaha/article/details/73194763
***********************/

#ifndef  KF_H
#define  KF_H

#include <Eigen/Dense>
using namespace Eigen;

// 卡尔曼滤波器类
class Kalman{

public:
 /*
 @breif Standard Kalman filter
 */
 MatrixXf &predict(); 

 /*
 @brief AFK: "Tracking Touched Trajectory on Capacitive Touch Panels Using Adjustable Weighted Prediction Covariance Matrix."
 @param  vx Velocity in x-axis direction
 @param  vy Velocity in y-axis direction
 @param  yetax  Error in x-axis direction :z(k)-H*x'(k)
 @param  yetay  Error in y-axis direction :z(k)-H*x'(k)
 */
 MatrixXf &AK_predict(double &vx, double &vy, double &yetax, double &yetay); 

 /*
 @brief Adactive velocit adjust 
 @param  vx Velocity in x-axis direction
 @param  vy Velocity in y-axis direction
 @param  yetax  Error in x-axis direction(innovation sequence):z(k)-H*x'(k)
 @param  yetay  Error in y-axis direction(innovation sequence):z(k)-H*x'(k)                              
 @V_L Low threshold of speed
 @V_H High threshold of speed
 */
 MatrixXf &AKV_predict(double vx, double vy, double yetax, double yetay, double V_L,double V_H); 
 
 /*
 @brief  STFK: "Position stimation and Smooth Tracking with a Fuzzy Logic-Based Adaptive Strong Tracking Kalman Filter for Capacitive Touch Panels."
 @param  measureParams Dimensionality of the measurement.
 @param  vx Velocity in x-axis direction
 @param  vy Velocity in y-axis direction
 @param  yetax  Error in x-axis direction() :z(k)-H*x'(k)
 @param  yetay  Error in y-axis direction :z(k)-H*x'(k)
 */
 MatrixXf &STFK_predict(int measureParams, double &vx, double &vy, double &yetax, double &yetay);

 // 根据测量值更新预测状态
 MatrixXf &correct(MatrixXf &measurement);

 // 根据测量值更新预测状态
 MatrixXf &AKV_correct(MatrixXf measurement,double vx, double vy);

 /* 
 @brief Re-initializes Kalman filter.
 @param dynamParams Dimensionality of the state.
 @param measureParams Dimensionality of the measurement.
 @param controlParams Dimensionality of the control vector.
 */
 void init(int DP, int MP, int CP);

 MatrixXf statePre; //预测状态 (x'(k)): x(k)=A*x(k-1)+B*u(k)
 MatrixXf stateOpt; //修正状态 (x(k)): x(k)=x'(k)+K(k)*(z(k)-H*x'(k))
 MatrixXf transMat; //状态转移矩阵 (A)
 MatrixXf transMat_slow; //状态转移矩阵 (A)
 MatrixXf contrMat; // 控制矩阵 (B) (未使用时为0)
 MatrixXf measureMat; // 测量矩阵 (H)
 MatrixXf processNoiseCov; // 过程噪声协方差矩阵 (Q)
 MatrixXf processNoiseCov_Slow; // 过程噪声协方差矩阵 (Q)  低速
 MatrixXf processNoiseCov_Fast; // 过程噪声协方差矩阵 (Q)  高速
 MatrixXf measureNosiseCov; // 测量噪声协方差矩阵 (R)
 MatrixXf measureNosiseCov_Slow; // 测量噪声协方差矩阵 (R)  低速
 MatrixXf measureNosiseCov_Fast; // 测量噪声协方差矩阵 (R)  高速
 MatrixXf errorCovpre; // 先验误差估计协方差矩阵 (P'(k)): P'(k)=A*P(k-1)*At + Q)
 MatrixXf Kgain;        // 卡尔曼增益矩阵 (K(k)): K(k)=P'(k)*Ht*inv(H*P'(k)*Ht+R)
 MatrixXf errorCovOpt;  // 后验误差估计协方差矩阵 (P(k)): P(k)=(I-K(k)*H)*P'(k)

 // STKF相关变量
 MatrixXf Vk;   
 MatrixXf Nk;
 MatrixXf Mk;
 double rho; // 忘却因子
 double get_lamda(int MP); // 计算次优缩放因子
 double lamda; // 次优缩放因子

 // AKF相关变量
 double get_beta(double vx, double vy, double yetax, double yetay); // 计算beta
 double get_beta_plus(double vx, double vy); // 计算beta
 double eps;  // 经验参数，取值为0.01
 double beta; // 一种权重，用于调整AKF的预测协方差矩阵

 // 临时矩阵
 MatrixXf tmp1;
 MatrixXf tmp2;
 MatrixXf tmp3;
 MatrixXf tmp4;
 MatrixXf tmp5;   // 创新序列 vk

private:
 // STKF私有成员函数
 MatrixXf &get_Vk();
 MatrixXf &get_Mk();
 MatrixXf &get_Nk();
};

#endif